--- Configs
-- @author Enno Sylvester
-- @copyright 2017, 2018, 2019, 2023
-- @module eatglobal.configs

local historie = [[
	Version 1.6:
	 - Added: Added Fields "__debug__" and "historie" to result-table.
	 - Added: Modulfähigkeit (Erweiterungsmöglichkeiten durch Module) implementiert.
	 - Changed: Switch to new system of initialization eatglopal-API (postLoadFn).
]]

local verionsData = {
	majorVersion = 1,
	minorVersion = 6,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local fileutils
local userSettings
local getValue
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		fileutils = global.fileutils
		userSettings = global.userSettings
		getValue = global.misc.getValue
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------


local function init ()
	if type(eat1963_global_data) ~= "table" then
		eat1963_global_data = {}
	end
	if type(eat1963_global_data.configs) ~= "table" then
		eat1963_global_data.configs = {}
	end
	return true
end

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	createConfig
	--	
	------------------------------------------------------------------------------------
	createConfig = function(id, params)
		local function checkLogFn (id)
			local config = eat1963_global_data.configs[id]
			if config.env.debug then
				config.logFn = global.utils.createLogFn (id)
				config.logFnCounter = global.utils.createLogFnCounter (id)
			else
				config.logFn = function (...) end
				config.logFnCounter = function (...) end
			end
		end
		local function createSetDebugFn ( id )
			local id = id
			local function fn ( _debug )
				eat1963_global_data.configs[id].env.debug = _debug
				checkLogFn (id)
			end
			return fn
		end
		local function createGetIdFn (id)
			local id = id
			local function fn ()
				return id
			end
			return fn
		end
		
		local function createSaveModSettingsFn (id)
			local id = id
			local function fn ()
				fileutils.saveData(eat1963_global_data.configs[id].modSettings, eat1963_global_data.configs[id].env.mod_settingsFile)
			end
			return fn
		end
		
		local function createSaveUserConfigFn (id)
			local id = id
			local function fn ()
				userSettings.flushSettings (id)
			end
			return fn
		end
		
		local function createImportUserConfigFn (id)
			local id = id
			local function fn (userConfig)
				eat1963_global_data.configs[id].userConfig = 
					global.utils.proxyTable(userSettings.importUserSettings(id, userConfig))
			end
			return fn
		end
		
		local function createInitializeUserConfigFn (id)
			local id = id
			local function fn (defaultUserSettings)
				eat1963_global_data.configs[id].userConfig = 
					global.utils.proxyTable(userSettings.create(id, defaultUserSettings))
			end
			return fn
		end
		
		if (init () and (type(id) == "string")) then
			if not eat1963_global_data.configs[id] then
				local temp = params or {}
				if (temp.env == nil) then
					temp.env = {}
				end
				if (temp.userConfig == nil) then
					temp.userConfig = {}
				end
				if (temp.modSettings == nil) then
					temp.modSettings = {}
				end
				if (temp.modifierSequenceList == nil) then
					temp.modifierSequenceList = {
						["01"] = id,
					}
				end
				-- Umgebung (env(ironment))
				temp.env.modPath = fileutils.getModPath(1)
				temp.env.settingsFile = temp.env.modPath.."settings.lua"
				temp.env.mod_settingsFile = temp.env.modPath.."mod_settings.lua"
				temp.env.scriptsPath = temp.env.modPath.."res/scripts/"
				temp.env.isWinOS = (os.getenv("OS") == "Windows_NT")
				
				temp.setDebug = createSetDebugFn(id)
				temp.getId = createGetIdFn(id)
				-- Mod-Settings
				temp.modSettings = global.utils.readDataEx(temp.env.mod_settingsFile)
				temp.saveModSettings = createSaveModSettingsFn (id)
				-- User-Settings
				temp.initializeUserConfig = createInitializeUserConfigFn (id)
				temp.saveUserConfig = createSaveUserConfigFn (id)
				temp.importUserConfig = createImportUserConfigFn (id)
				-- zuweisen der Daten
				eat1963_global_data.configs[id] = temp
				
				--
				checkLogFn(id)
			end
			return eat1963_global_data.configs[id]
		else
			return {}
		end
	end,
	------------------------------------------------------------------------------------
	--	getConfig
	--	
	------------------------------------------------------------------------------------
	getConfig = function(id)
		init()
		return getValue(eat1963_global_data.configs[id], {})
	end,
	------------------------------------------------------------------------------------
	--	existsConfig
	--	
	------------------------------------------------------------------------------------
	existsConfig = function(id)
		return (init() and (type(eat1963_global_data.configs[id]) == "table"))
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
