--- math
-- @author Enno Sylvester
-- @copyright 2023
-- @module eatglobal.math

local historie = [[
	Version 1.1:
	 - Release
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 1,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
local randomTable

functionClass = {
	------------------------------------------------------------------------------------
	--	addPercent
	--	Addiert percent-Prozent zu arg hinzu.
	------------------------------------------------------------------------------------
	addPercent = function(arg, percent) return arg + ((arg * percent) / 100) end,
	------------------------------------------------------------------------------------
	--	calcScale
	--	
	------------------------------------------------------------------------------------
	calcScale = function(dist, angle)
		if (angle < .001) then
			return dist
		end
				
		local pi2 = math.pi / 2	
		local sqrt2 = math.sqrt(2)
				
		local scale = 1.0
		if (angle >= pi2) then 
			scale = 1.0 + (sqrt2 - 1.0) * ((angle - pi2) / pi2)
		end

		return .5 * dist / math.cos(.5 * math.pi - .5 * angle) * angle * scale
	end,
	------------------------------------------------------------------------------------
	--	clamp
	--	Tariert value in den Bereich zwischen min und max.
	------------------------------------------------------------------------------------
	clamp = function(value, min, max) return math.min(math.max(value, min), max) end,
	------------------------------------------------------------------------------------
	--	clearNumberArray
	--	Macht letztlich genau das Gleiche, wie round.
	--	Unterschied: Wenn 'decimals' nicht valide, so werden 3 statt 0 Nachkommastellen
	--	als Standard gesetzt.
	------------------------------------------------------------------------------------
	clearNumberArray = function(numbers, decimals)
		return functionClass.round(numbers, ((type(decimals) ~= "number") and 3 or decimals))
	end,
	------------------------------------------------------------------------------------
	--	makeMiddledRange
	--	Erzeugt ein Array aus Zahlen, welches von 0 ausgehend sowohl negativ als auch
	--	positiv bis 'max' reicht, im Abstand von 'step'.
	--	Das Ergebnis wird auf 'decimals' Stellen nach dem Komma gerundet.
	--	Ist 'decimals' nicht valide so wird auf 3 Stellen nach dem Komma gerundet.
	------------------------------------------------------------------------------------
	makeMiddledRange= function(max, step, decimals)
		local steps = functionClass.makeRange(0, math.abs(max), step, decimals)
		local result = eatSetEx.new({0})
		for i = 2, #steps do
			table.insert(result, 1, -steps[i])
			table.insert(result, steps[i])
		end
		return result
	end,
	------------------------------------------------------------------------------------
	--	makeRange
	--	Erzeugt ein Array aus Zahlen von 'min' bis 'max' im Abstand von 'step'.
	--	Das Ergebnis wird auf 'decimals' Stellen nach dem Komma gerundet.
	--	Ist 'decimals' nicht valide so wird auf 3 Stellen nach dem Komma gerundet.
	------------------------------------------------------------------------------------
	makeRange = function(min, max, step, decimals)
		return functionClass.splitNumberRange(min, max, functionClass.round(math.abs((max - min) / math.abs(step))), decimals)
	end,
	------------------------------------------------------------------------------------
	--	random
	--	Alternative zu "math.random", welche "zufälligere" Zufallszahlen liefert.
	--	Inspiration von http://lua-users.org/wiki/MathLibraryTutorial
	--	Abschnitt math.random , math.randomseed
	--	Rückgaben, wie bei "math.random":
	--	lower und upper nicht angegeben: reelle Zahl zwischen 0 und 1
	--	nur lower angegeben: integer zwischen 1 und lower (beide Inclusive)
	--	lower und upper angegeben: integer zwischen lower und upper (beide Inclusive)
	------------------------------------------------------------------------------------
	random = function(lower, upper)
		if randomTable == nil then
			randomTable = {}
			for i = 1, 97 do
				randomTable[i] = math.random()
			end
		end
		local x = math.random()
		local i = 1 + math.floor(97 * x)
		x, randomTable[i] = randomTable[i], x
		
		if (type(lower) == "number") then
			if  (type(upper) == "number") then
				if (lower > upper) then
					lower, upper = upper, lower
				end
				return functionClass.clamp(functionClass.round((x * (upper - lower)) + lower), lower, upper)
			else
				return functionClass.clamp(functionClass.round(x * lower), 1, lower)
			end
		else
			return x
		end
	end,
	------------------------------------------------------------------------------------
	--	randomDeg
	--	Liefert einen zufälligen Wert zwischen -Deg/2 und +Deg/2
	--	Möchte man z.B. ein Modell um die Z-Achse zufällig um +/- 22.5° drehen:
	--	Z-Achse = randomDeg(45)
	------------------------------------------------------------------------------------
	randomDeg = function(Deg) return functionClass.random(Deg) - (Deg / 2) end,
	------------------------------------------------------------------------------------
	--	round
	--	Rundet (kaufmännisch) 'var' auf 'decimals' Stellen nach dem Komma.
	--	Ist decimals ungültig (nil) so wird als Default 0 genommen.
	--	'var' kann eine Zahl, ein Nummernstring oder ein Array aus Zahlen sein.
	------------------------------------------------------------------------------------
	round = function(var, decimals)
		local factor = 10 ^ ((type(decimals) ~= "number") and 0 or math.abs(decimals))
		------------------------------------------------------------
		local function doRound(n)
			return math.floor((n * factor) + 0.5) / factor
		end
		local function _do(value)
			if (type(value) == "number") then
				return doRound(value)
			elseif (type(value) == "string") then
				return doRound(tonumber(value) or 0)
			else
				return 0
			end
		end
		------------------------------------------------------------
		if (type(var) == "table") then
			local result = eatSetEx.new()
			for i, value in ipairs(var) do
				result:insert(_do(value))
			end
			return result
		else
			return _do(var)
		end
	end,
	------------------------------------------------------------------------------------
	--	sign
	--	Liefert die Richtung von x.
	------------------------------------------------------------------------------------
	sign = function(x) return (x > 0) and 1 or (x < 0) and -1 or 0 end,
	------------------------------------------------------------------------------------
	--	splitNumberRange
	--	Erzeugt ein Array aus Zahlen mit der Anzahl 'steps' von 'first' bis 'last'.
	--	Das Ergebnis wird auf 'decimals' Stellen nach dem Komma gerundet.
	--	Ist 'decimals' nicht valide so wird auf 3 Stellen nach dem Komma gerundet.
	------------------------------------------------------------------------------------
	splitNumberRange = function(first, last, steps, decimals)
		steps = math.abs(steps)
		local step = (last - first) / steps
		local result = {}
		for i = 0, steps do
			result[#result + 1] = first + (i * step)
		end
		return functionClass.clearNumberArray(result, decimals)
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB, math}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
