--- Mod-Tools
-- @author Enno Sylvester
-- @copyright 2019, 2020, 2023
-- @module eatglobal.mod

local historie = [[
	Version 1.3:
	 - Changed: "__debug__" und "historie" hinzugefügt.
	Version 1.4:
	 - Changed: Switch to new system of initialization eatglopal-API (postLoadFn).
	Version 1.5:
	 - Added: Support für "priority"-Parameter bei Modifiern implementiert.
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 5,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local table
local getValue
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		table = global.table
		getValue = global.misc.getValue
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass
local stdPriority = 10

------------------------------------------------------------------------------------
--	locale Daten
------------------------------------------------------------------------------------
local subKeyData = {
	["model/vehicle"] = "loadModel",
	["model/person"] = "loadModel",
	["model/car"] = "loadModel",
	["model/rock"] = "loadModel",
	["model/tree"] = "loadModel",
	["model/signal"] = "loadModel",
	["model/other"] = "loadModel",

	multipleUnit = "loadMultipleUnit",
	street = "loadStreet",
	track = "loadTrack",
	bridge = "loadBridge",
	tunnel = "loadTunnel",
	railroadCrossing = "loadRailroadCrossing",
	trafficLight = "loadTrafficLight",
	environment = "loadEnvironment",
	construction = "loadConstruction",
	module = "loadModule",
	autoGroundTex = "autoGroundTex",
	groundTex = "loadGroundTex",
	terrainGenerator = "loadTerrainGenerator",
	terrainMaterial = "loadTerrainMaterial",
	cargoType = "loadCargoType",
	grass = "loadGrass",
	gameScript = "loadGameScript",
	climate = "loadClimate",
}

------------------------------------------------------------------------------------
--	Tool(s)
------------------------------------------------------------------------------------
local function baseTable()
	return eat1963_global_data.mod_lua.filefilters
end

local function isHideObject(fileName, tbl)
	for i = 1, #tbl do
		if string.ends(fileName, tbl[i]) then
			return true
		end
	end
	
	return false
end

local function getTable(tbl, key)
	return global.misc.getValue(tbl[key], {})
end

local function handleHideObjectsFF(fileName, data, key)
	return (not isHideObject(fileName, baseTable().hideObjects.ffObjects))
end

local function handleHideObjectsMF(fileName, data, key)
	local function doIt(availability)
		availability.yearTo = 1849
		availability.yearFrom = 1849
	end
	if isHideObject(fileName, baseTable().hideObjects.mfObjects) then
		if (type(data.availability) == "table") then
			doIt(data.availability)
		elseif ((type(data.metadata) == "table") and (type(data.metadata.availability) == "table")) then
			doIt(data.metadata.availability)
		else
			doIt(data)
		end
	end
	return data
end

------------------------------------------------------------------------------------
--	doUpdateFilter
--	
------------------------------------------------------------------------------------
local updateRunningTables = false
local function doUpdateRunningTables()
	------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------
	local function doUpdateRunningTable(tbl)
		local updateTable = {}
		for key, values in pairs(baseTable()[tbl]) do
			local result = eatSetEx.new()
			for i, value in ipairs(values) do
				if ((type(value) == "table") and (type(value.runFn) == "function")) then
					result:insert(value)
				elseif (type(value) == "function") then
					result:insert({
						runFn = value,
						priority = stdPriority,
					})
				end
			end
			updateTable[key] = result
		end
		baseTable()[tbl] = updateTable
	end
	------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------
	if (not updateRunningTables) then
		doUpdateRunningTable("filters")
		doUpdateRunningTable("filterHooks")
		doUpdateRunningTable("modifier")
		doUpdateRunningTable("modifierHooks")
		updateRunningTables = true
	end
	
end

------------------------------------------------------------------------------------
--	newApplyFileFilters
--	
------------------------------------------------------------------------------------
local function newApplyFileFilters(key, fileName, data)
	------------------------------------------------------------------------------------
	--	getKey
	--	
	------------------------------------------------------------------------------------
	local function getKey(t)
		if string.ends(key, "/") then
			for k, v in pairs(t) do
				if string.starts(k, key) then
					return k
				end
			end
		end
		return key
	end
	
	------------------------------------------------------------------------------------
	--	doHook
	--	
	------------------------------------------------------------------------------------
	local function doHook(result)
		local _key = getKey(baseTable().filterHooks)
		local hooks = eatSetContainer(getTable(baseTable().filterHooks, "all")):union(getTable(baseTable().filterHooks, _key))
		if (#hooks > 0) then
			local _data = table.copyTable(data)
			for hook in hooks:_pairs() do
				hook.runFn(fileName, _data, _key, result)
			end
		end
	end
	
	------------------------------------------------------------------------------------
	--	doFilter
	--	
	------------------------------------------------------------------------------------
	local function doFilter()
		local _key = getKey(baseTable().filters)
		local filters = eatSetContainer(getTable(baseTable().filters, "all")):union(getTable(baseTable().filters, _key))
		logFn("filters = ", filters)
		for filter in filters:_pairs() do
			if not filter.runFn(fileName, data, _key) then
				return nil
			end
		end
		
		return data
	end
	
	------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------
	doUpdateRunningTables()
		
	local result = baseTable().origApplyFileFilters(key, fileName, data)
	if result then
		result = doFilter()
	end
	doHook(result)

	return result
end

------------------------------------------------------------------------------------
--	newApplyModifiers
--	
------------------------------------------------------------------------------------
local function newApplyModifiers(key, fileName, data)
	------------------------------------------------------------------------------------
	--	doHook
	--	
	------------------------------------------------------------------------------------
	local function doHook(result)
		local hooks = eatSetContainer(getTable(baseTable().modifierHooks, "all")):union(getTable(baseTable().modifierHooks, key))
		if (#hooks > 0) then
			local _data = table.copyTable(result)
			for hook in hooks:_pairs() do
				hook.runFn(fileName, _data, key)
			end
		end
	end
	
	------------------------------------------------------------------------------------
	--	doModifier
	--	
	------------------------------------------------------------------------------------
	local function doModifier(result)
		local modifiers = eatSetContainer(getTable(baseTable().modifier, "all")):union(getTable(baseTable().modifier, key))

		for modifier in modifiers:_pairs(function(a, b) return (getValue(a.priority, stdPriority) < getValue(b.priority, stdPriority)) end) do
			result = modifier.runFn(fileName, result, key)
		end
		
		return result
	end
	
	------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------
	doUpdateRunningTables()
		
	data = baseTable().origApplyModifiers(key, fileName, data)
	data = doModifier(data)
	doHook(data)
	
	return data
end

------------------------------------------------------------------------------------
--	init
--	
------------------------------------------------------------------------------------
local function init ()
	-- globale Daten
	if (type(eat1963_global_data) ~= "table") then
		eat1963_global_data = {}
	end
	-- mod_lua
	if (type(eat1963_global_data.mod_lua) ~= "table") then
		eat1963_global_data.mod_lua = {
			version = versionsData,
		}
	end
	-- globale Filter und Modifier
	if (type(eat1963_global_data.mod_lua.filefilters) ~= "table") then
		eat1963_global_data.mod_lua.filefilters = {
			filters = {},
			modifier = {},
			filterHooks = {},
			modifierHooks = {},
			hideObjects = {
				ffRegistered = false,
				mfRegistered = false,
				ffObjects = {},
				mfObjects = {},
			},
			origApplyFileFilters = applyFileFilters,
			origApplyModifiers = applyModifiers,
		}
		applyFileFilters = newApplyFileFilters
		applyModifiers = newApplyModifiers
	else
		if (eatVersion.new(eat1963_global_data.mod_lua.version) < versionsData) then
			eat1963_global_data.mod_lua.version = versionsData
			applyFileFilters = newApplyFileFilters
			applyModifiers = newApplyModifiers
		end
	end
end

local function doInsert(tbl, key, runFn, priority)
	local t = getTable(tbl, key)
	table.insert(t, {
		runFn = runFn,
		priority = getValue(priority, stdPriority),
	})
	tbl[key] = t
end

------------------------------------------------------------------------------------
--	functionClass
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	addFileFilter
	--	
	------------------------------------------------------------------------------------
	addFileFilter = function(key, runFn)
		init()
		doInsert(baseTable().filters, key, runFn)
	end,
	------------------------------------------------------------------------------------
	--	addFilterHook
	--	
	------------------------------------------------------------------------------------
	addFilterHook = function(key, runFn)
		init()
		doInsert(baseTable().filterHooks, key, runFn)
	end,
	------------------------------------------------------------------------------------
	--	addModifier
	--	
	------------------------------------------------------------------------------------
	addModifier = function(key, runFn, priority)
		init()
		doInsert(baseTable().modifier, key, runFn, priority)
	end,
	------------------------------------------------------------------------------------
	--	addModifierHook
	--	
	------------------------------------------------------------------------------------
	addModifierHook = function(key, runFn)
		init()
		doInsert(baseTable().modifierHooks, key, runFn)
	end,
	------------------------------------------------------------------------------------
	--	addCommonApiSavedModifier
	--	
	------------------------------------------------------------------------------------
	addCommonApiSavedModifier = function(key, runFn, priority)
		local function doFilter(fileName, data, _key)
			runFn(fileName, data, _key)
			return true
		end
		
		functionClass.addFileFilter(key, doFilter)
		functionClass.addModifier(key, runFn, priority)
	end,
	------------------------------------------------------------------------------------
	--	getSubKey
	--	
	------------------------------------------------------------------------------------
	getSubKey = function(key)
		for k, v in pairs(subKeyData) do
			if string.starts(k, key) then
				return v
			end
		end

		return key
	end,
	------------------------------------------------------------------------------------
	--	hideObject
	--	
	------------------------------------------------------------------------------------
	hideObject = function(data, useModifier)
		local function doHide(name)
			useModifier = (useModifier == nil) and false or useModifier
			local hideObjects
			if useModifier then
				if (not baseTable().hideObjects.mfRegistered) then
					functionClass.addCommonApiSavedModifier("all", handleHideObjectsMF)
					baseTable().hideObjects.mfRegistered = true
				end
				hideObjects = baseTable().hideObjects.mfObjects
			else
				if (not baseTable().hideObjects.ffRegistered) then
					functionClass.addFileFilter("all", handleHideObjectsFF)
					baseTable().hideObjects.ffRegistered = true
				end
				hideObjects = baseTable().hideObjects.ffObjects
			end
			
			hideObjects[#hideObjects + 1] = name
		end
		-----------------------------------------
		init()
		if (type(data) == "table") then
			for name in eatSetContainer(data):_pairs() do
				if global.misc.isValidId(name) then
					doHide(name)
				end
			end
		elseif global.misc.isValidId(data) then
			doHide(data)
		end
	end,
}

local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
