--- table
-- @author Enno Sylvester
-- @copyright 2023
-- @module eatglobal.table

local historie = [[
	Version 1.1:
	 - Release
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 1,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	copyTable
	--	Erzeugt eine Kopie von 'tbl'.
	------------------------------------------------------------------------------------
	copyTable = function(tbl)
		local function doCopy(t)
			if (type(t) == 'table') then
				local result = {}
				for key, value in next, t, nil do
					result[doCopy(key)] = doCopy(value)
				end
				return setmetatable(result, doCopy(getmetatable(t)))
			else -- number, string, boolean, etc
				return t
			end
		end
		return doCopy(tbl)
	end,
	------------------------------------------------------------------------------------
	--	concatTable
	--	Fügt alle Einträge aus 'tbl' zusammen und gibt sie als String aus.
	--	Ist 'seperator' ein gültiger String (z.B. "/n" oder ""t"), wird er zwischen den
	--	Einträgen der Tabelle eingefügt.
	------------------------------------------------------------------------------------
	concatTable = function(tbl, seperator)
		local stack = {}
		for i, value in ipairs(tbl) do
			if ((i > 1) and global.misc.isValidId(seperator)) then
				table.insert(stack, seperator)
			end
			table.insert(stack, (type(value) == "string") and value or tostring(value))
		end
		return table.concat(stack)
	end,
	------------------------------------------------------------------------------------
	--	sortPairs
	--	Stellt eine Funktion zum sortierten Iterieren einer Tabelle zur Verfügung.
	--	Ist 'orderFn' gültig, so wird diese zum Sortieren genutzt; andernfalls
	--	Standardsortierung mit 'table.sort'.
	------------------------------------------------------------------------------------
	sortPairs= function(tbl, orderFn)
		local keys = {}
		for k in pairs(tbl) do
			keys[#keys + 1] = k
		end

		if orderFn then
			table.sort(keys, function(a,b) return orderFn(tbl, a, b) end)
		else
			table.sort(keys)
		end

		local i = 0
		return function()
			i = i + 1
			if keys[i] then
				return keys[i], tbl[keys[i]]
			end
		end
	end,
	------------------------------------------------------------------------------------
	--	isArray
	--	True, wenn 'tbl' ein Array ist.
	------------------------------------------------------------------------------------
	isArray= function(tbl)
		local keys = {}
		for i in ipairs(tbl) do
			keys[i] = true
		end
		for key in pairs(tbl) do
			if not keys[key] then
				return false
			end
		end
		return true
	end,
	------------------------------------------------------------------------------------
	--	unpack
	--	Da lua --> unpack wohl nicht mehr verfügbar, hier eine Eigene.
	------------------------------------------------------------------------------------
	unpack = function(tbl)
		local function doIt(_tbl, i)
			local value = _tbl[i]
			if (value ~= nil) then
				return value, doIt(_tbl, i + 1)
			end
		end
		
		if (type(tbl) == "table") then
			return doIt(tbl, 1)
		end
	end,
	------------------------------------------------------------------------------------
	--	hasMetaFunction
	--	Prüft, ob in Tabelle tbl eine oder mehrere Metafunktione vorhanden sind
	--	@param tbl:	Tabelle, welche durchsucht wird.
	--	@param fn:	Name der gesuchten Funktion (String) oder Tabelle mit Namen (Strings)
	--							der gesuchten Funktionen
	--	@return:		true, wenn vorhanden; andernfalls false
	--	@use:				if hasMetaFunction(tbl, "getValue1") then ... /
	--							if hasMetaFunction(tbl, {"getValue1", "getValue2", "getValue2", ...}) then ...
	------------------------------------------------------------------------------------
	hasMetaFunction = function(tbl, fn)
		if (type(tbl) == "table") then
			local m = getmetatable(tbl) or {}
			if global.misc.isValidId(fn) then
				return (type(m[fn]) == "function")
			elseif ((type(fn) == "table") and (#fn > 0)) then
				for i, value in ipairs(fn) do
					if ((not global.misc.isValidId(value)) or (type(m[value]) ~= "function")) then
						return false
					end
				end
				return true
			end
		end
		return false
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB, table}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
