--- vec3
-- @author Enno Sylvester
-- @copyright 2018, 2021
-- @module eatglobal.vec3

local historie = [[
	Version 1.1:
	 - Release
	Version 1.2:
	 - Added: Added Fields "__debug__" and "version" to result-table vec3.
	 - Added: Functions angleUnit, calcMax, calcMin, calcNSectors, calcSteps, calcZArgAndAngel, copy, cross, degToRad, distance, getMiddle,
						getZAngel, getZData, dot, length, lengthTo, max, normalize, projectZPoint, rotateX, rotateY, rotateZ, split, toParams, 
						toArray and xyAngle now are Class-Function.
	 - Changed: vec3.new now accepts arr3, vec3 and numbers as parameters.
	 - Changed: +, -, * and / now accept both vec3 and numbers.
	 - Changed: Functions with table returns aka lists, returns now as eatSetEx.
	 - Changed: Deletion of all import functions, because vec3.new covers all possibilities.
	 - Changed: Deletion of the functions vec3.copy and vec3.mt, because vec3.new covers these functionalities. Class function copy is retained.
]]

local verionsData = {
	majorVersion = 1,
	minorVersion = 2,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local arr3
local math
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		arr3 = global.arr3
		math = global.math
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local function getVec3(x, y, z)
	if (type(x) == "table") then
		if ((type(x[1]) == "number") and (type(x[2]) == "number")) then
			-- x is a arr2 or arr3
			return {x = x[1], y = x[2], z = (type(x[3]) == "number") and x[3] or 0}
		elseif ((type(x.x) == "number") and (type(x.y) == "number")) then
			-- x is a vec2, vec3 or vec4
			return {x = x.x, y = x.y, z = (type(x.z) == "number") and x.z or 0}
		end
	elseif (type(x) == "number") then
		-- x is a number; We look to see if y and z are also numbers. Otherwise they get x as value
		return {x = x, y = (type(y) == "number") and y or x, z = (type(z) == "number") and z or x}
	end
	-- nothing; return default
	return {x = 0, y = 0, z = 0}
end

local function angleArg(arg)
	return math.acos(math.clamp(arg, -1, 1))
end

local function getSinCos(rot)
	return math.sin(rot), math.cos(rot)
end

local mt
mt = {
  __add = function (self, rhs) -- "add" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x + rhs, y = self.y + rhs, z = self.z + rhs}, mt)
		else
			local v = getVec3(rhs)
			return setmetatable({x = self.x + v.x, y = self.y + v.y, z = self.z + v.z}, mt)
			--return functionClass.new({x = self.x + v.x, y = self.y + v.y, z = self.z + v.z})
		end
  end,
	__sub = function (self, rhs) -- "sub" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x - rhs, y = self.y - rhs, z = self.z - rhs}, mt)
		else
			local v = getVec3(rhs)
			return setmetatable({x = self.x - v.x, y = self.y - v.y, z = self.z - v.z}, mt)
		end
  end,
	__mul = function (self, rhs) -- "mul" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x * rhs, y = self.y * rhs, z = self.z * rhs}, mt)
		else
			local v = getVec3(rhs)
			return setmetatable({x = self.x * v.x, y = self.y * v.y, z = self.z * v.z}, mt)
		end
	end,
	__div = function (self, rhs) -- "div" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x / rhs, y = self.y / rhs, z = self.z / rhs}, mt)
		else
			local v = getVec3(rhs)
			return setmetatable({x = self.x / v.x, y = self.y / v.y, z = self.z / v.z}, mt)
		end
	end,
	__index = {
		angleUnit = function(self, vec)
			return angelArg(self:dot(vec))
		end,
		calcMax = function(self, vec)
			return setmetatable({x = math.max(self.x, vec.x), y = math.max(self.y, vec.y), z = math.max(self.z, vec.z)}, mt)
		end,
		calcMin = function(self, vec)
			return setmetatable({x = math.min(self.x, vec.x), y = math.min(self.y, vec.y), z = math.min(self.z, vec.z)}, mt)
		end,
		calcNSectors = function(self, vec, countSectors)
			local result = eatSetEx.new({self:copy()})
			local vecEnd = functionClass.new(vec)
			local vecSector = (vecEnd - self) / countSectors
			for i = 1, countSectors - 1 do
				result:insert(self + (vecSector * i))
			end
			result:insert(vecEnd)
			
			return result
		end,
		calcSteps = function(self, vec, range)
			return math.round(math.max((self - vec):max(), range) / range)
		end,
		calcZArgAndAngel = function(self, vec, radius)
			local arg, angel, length = self:getZData(vec, radius)
			return arg, angel * length
		end,
		copy = function(self)
			local x, y, z = self:toParams()
			return setmetatable({x = x, y = y, z = z}, mt)
		end,
		cross = function(self, vec)
			return setmetatable({
				x = self.y * vec.z - self.z * vec.y,
				y = self.z * vec.x - self.x * vec.z,
				z = self.x * vec.y - self.y * vec.x
			}, mt)
		end,
		degToRad = function(self)
			return setmetatable({x = math.rad(self.x), y = math.rad(self.y), z = math.rad(self.z)}, mt)
		end,
		distance = function(self, vec)
			return (self - vec):length()
		end,
		getMiddle = function(self, vec)
			return self + ((functionClass.new(vec) - self) / 2)
		end,
		getZAngel = function(self, vec)
			local function calcTotalLength(yLength, z)
				local fac = (yLength < 0) and -1 or 1
				return math.sqrt(yLength^2 + z^2) * fac
			end
			
			local sx, sy, sz = self:toParams()
			local ex, ey, ez = functionClass.new(vec):toParams()
			local z = math.abs(ez - sz)
			return z / calcTotalLength(ey - sy, z)
		end,
		getZData = function(self, vec, radius)
			local arg = vec.z - self.z
			return arg, angleArg(arg), self:lengthTo(vec, radius)
		end,
		dot = function(self, vec)
			return self.x * vec.x + self.y * vec.y + self.z * vec.z
		end,
		length = function(self)
			return math.sqrt(self.x^2 + self.y^2 + self.z^2)
		end,
		lengthTo = function(self, vec, radius)
			local r = radius or 0
			local l = math.abs(self:distance(vec))
			if (r == 0) then
				return l
			else
				return 2 * r * math.asin( l / (2 * r) )
			end
		end,
		max = function(self)
			local result = 0
			for k, value in pairs(self) do
				result = math.max(result, math.abs(value))
			end
			return result
		end,
		normalize = function(self)
			return self * (1 / self:length())
		end,
		projectZPoint = function(self, angel, yLength)
			return setmetatable({x = self.x, y = self.y + yLength, z = self.z + (math.tan(angel) * yLength)}, mt)
		end,
		--- alle rotate...-Funktionen mit rot in RAD
		-- Umrechnung: aRAD = math.rad(aDEG)
		rotateX = function(self, rot)
			local s, c = getSinCos(rot)
			return setmetatable({x = self.x, y = (self.y * c - self.z * s), z = (self.y * s + self.z * c)}, mt)
		end,
		rotateY= function(self, rot)
			local s, c = getSinCos(rot)
			return setmetatable({x = (self.x * c + self.z * s), y = self.y, z = (self.x * -s + self.z * c)}, mt)
		end,
		rotateZ= function(self, rot)
			local s, c = getSinCos(rot)
			return setmetatable({x = (self.x * c - self.y * s), y = (self.y * s + self.y * c), z = self.z}, mt)
		end,
		split = function(self, vec, steps, insertLast)
			local result = eatSetEx.new({self:copy()})
			local vecEnd = functionClass.new(vec)
			local addVec = (vecEnd - self) / steps
			local currentVec = self:copy()
			for i = 1, steps - 1 do
				local nextVec = currentVec + addVec
				result:insert(nextVec)
				currentVec = nextVec
			end
			
			if ((type(insertLast) ~= "boolean") and true or insertLast) then
				result:insert(vecEnd)
			end
			
			return result
		end,
		swap = function(self)
			return setmetatable({self.z, self.y, self.x}, mt)
		end,
		toParams = function(self)
			return self.x, self.y, self.z
		end,
		toArray = function(self)
			return arr3.new(self)
		end,
		toString = function(self)
			return tostring(self)
		end,
		xyAngle = function(self)
			return math.atan2(self.y, self.x)
		end,
	},
	__tostring = function(self)
		return string.format("{x = %.4f, y = %.4f, z = %.4f}", self:toParams())
	end,
}

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	vec3.new(x, y, z):
	--	1.	x is a number. Then y and z can also be numbers. If not, they take the value of x.
	--	2.	x is a point (arr3)
	--	3.	x is an arr2, then z becomes 0
	--	4.	x is a vec3 / vec4
	--	5.	x is an vec2, then z becomes 0
	--	Otherwise returns vec3 with {x = 0, y = 0, z = 0}
	------------------------------------------------------------------------------------
	new = function(x, y, z)
		return setmetatable(getVec3(x, y, z), mt)
	end,
	------------------------------------------------------------------------------------
	--	mtVecs
	--	
	------------------------------------------------------------------------------------
	mtVecs = function(vecs)
		local result = eatSetEx.new()
		for i = 1, #vecs do
			result:insert(functionClass.new(vecs[i]))
		end
		return result
	end,
	------------------------------------------------------------------------------------
	--	add
	--	
	------------------------------------------------------------------------------------
	add = function(a, b)
		return functionClass.new(a) + b
	end,
	------------------------------------------------------------------------------------
	--	sub
	--	
	------------------------------------------------------------------------------------
	sub = function(a, b)
		return functionClass.new(a) - b
	end,
	------------------------------------------------------------------------------------
	--	mul
	--	
	------------------------------------------------------------------------------------
	mul = function(f, vec)
		return functionClass.new(vec) * f
	end,
	------------------------------------------------------------------------------------
	--	dot
	--	
	------------------------------------------------------------------------------------
	dot = function(a, b)
		return functionClass.new(a):dot(b)
	end,
	------------------------------------------------------------------------------------
	--	cross
	--	
	------------------------------------------------------------------------------------
	cross = function(a, b)
		return functionClass.new(a):cross(b)
	end,
	------------------------------------------------------------------------------------
	--	length
	--	
	------------------------------------------------------------------------------------
	length = function(vec)
		return functionClass.new(vec):length()
	end,
	------------------------------------------------------------------------------------
	--	distance
	--	
	------------------------------------------------------------------------------------
	distance = function(a, b)
		return functionClass.new(a):distance(b)
	end,
	------------------------------------------------------------------------------------
	--	max
	--	
	------------------------------------------------------------------------------------
	max = function(vec)
		return functionClass.new(vec):max()
	end,
	------------------------------------------------------------------------------------
	--	lengthBetween
	--	
	------------------------------------------------------------------------------------
	lengthBetween = function(vec0, vec1, radius)
		return functionClass.new(vec0):lengthTo(vec1, radius)
	end,
	------------------------------------------------------------------------------------
	--	getMiddle
	--	
	------------------------------------------------------------------------------------
	getMiddle = function(vec0, vec1)
		return functionClass.new(vec0):getMiddle(vec1)
	end,
	------------------------------------------------------------------------------------
	--	calcSteps
	--	
	------------------------------------------------------------------------------------
	calcSteps = function(vec0, vec1, range)
		return functionClass.new(vec0):calcSteps(vec1, range)
	end,
	------------------------------------------------------------------------------------
	--	split
	--	
	------------------------------------------------------------------------------------
	split = function(vec0, vec1, steps, insertLast)
		return functionClass.new(vec0):split(vec1, steps, insertLast)
	end,
	------------------------------------------------------------------------------------
	--	calcNSectors
	--	
	------------------------------------------------------------------------------------
	calcNSectors = function(vec0, vec1, countSectors)
		return functionClass.new(vec0):calcNSectors(vec1, countSectors)
	end,
	------------------------------------------------------------------------------------
	--	getZAngel
	--	
	------------------------------------------------------------------------------------
	getZAngel = function(pStart, pEnd)
		return functionClass.new(pStart):getZAngel(pEnd)
	end,
	------------------------------------------------------------------------------------
	--	projectZPoint
	--	
	------------------------------------------------------------------------------------
	projectZPoint = function(vec, angel, yLength)
		return functionClass.new(vec):projectZPoint(angel, yLength)
	end,
	------------------------------------------------------------------------------------
	--	normalize
	--	
	------------------------------------------------------------------------------------
	normalize = function(vec)
		return functionClass.new(vec):normalize()
	end,
	------------------------------------------------------------------------------------
	--	angleUnit
	--	
	------------------------------------------------------------------------------------
	angleUnit = function(a, b)
		return functionClass.new(a):angleUnit(b)
	end,
	------------------------------------------------------------------------------------
	--	xyAngle
	--	
	------------------------------------------------------------------------------------
	xyAngle = function(vec)
		return functionClass.new(vec):xyAngle()
	end,

	--- alle rotate...-Funktionen mit rot in RAD
	-- Umrechnung: aRAD = math.rad(aDEG)
	------------------------------------------------------------------------------------
	--	rotateX
	--	
	------------------------------------------------------------------------------------
	rotateX = function(vec, rot)
		return functionClass.new(vec):rotateX(rot)
	end,
	------------------------------------------------------------------------------------
	--	rotateY
	--	
	------------------------------------------------------------------------------------
	rotateY = function(vec, rot)
		return functionClass.new(vec):rotateY(rot)
	end,
	------------------------------------------------------------------------------------
	--	rotateZ
	--	
	------------------------------------------------------------------------------------
	rotateZ = function(vec, rot)
		return functionClass.new(vec):rotateZ(rot)
	end,

	----------------------

	------------------------------------------------------------------------------------
	--	degToRad
	--	
	------------------------------------------------------------------------------------
	degToRad = function(vec)
		return functionClass.new(vec):degToRad()
	end,
	------------------------------------------------------------------------------------
	--	getZData
	--	
	------------------------------------------------------------------------------------
	getZData = function(vec0, vec1, radius)
		return functionClass.new(vec0):getZData(vec1, radius)
	end,
	------------------------------------------------------------------------------------
	--	calcZArgAndAngel
	--	
	------------------------------------------------------------------------------------
	calcZArgAndAngel = function(vec0, vec1, radius)
		return functionClass.new(vec0):calcZArgAndAngel(vec1, radius)
	end,

	-- Exportfunktionen
	------------------------------------------------------------------------------------
	--	toArray
	--	
	------------------------------------------------------------------------------------
	toArray = function(vec)
		return arr3.new(vec)
	end,
	------------------------------------------------------------------------------------
	--	toParams
	--	
	------------------------------------------------------------------------------------
	toParams = function(vec)
		return vec.x, vec.y, vec.z
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

local vec3 = require "vec3"
return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB, vec3}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
